package com.ho.test.mail.ex1.server;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * <p>Title: FakeSMTPServer</p>
 * 
 * <p>
 * This class is used to simulate a mail server (SMTP).
 * <br/>It opens the 25 port, receipt the request and dump them to the console.
 * </p>
 * 
 * <p>
 * <b>usage : </b>
 * <br/>// create & directly run a new SMTP server on the default port 25 :
 * <br/>FakeSMTPServer smtpServer = new FakeSMTPServer();
 * <br>
 * <br/>// create & directly run a new SMTP server on the custom port 2342 :
 * <br/>FakeSMTPServer smtpServer = new FakeSMTPServer(2342)
 * <br/>
 * <br/>// Retrieve the full SMTP server log
 * <br/>String serverLog = smtpServer.getProtocolExchangeLog()
 * <br/>System.out.println("server log : \n" + serverLog);
 * <br/>
 * <br/>// Determine if a command had been executed in the SMTP server 
 * <br/>boolean isQuitCommandPresentInLog = smtpServer.isCommandPresentInLog("QUIT");
 * </p>
 * 
 */
public class FakeSMTPServer extends ServerSocket implements Runnable {

	// -------------------------------------------------------------- CONSTANTS
	/**
	 * The port number used by the SMTP server.
	 */
	public final static int DEFAULT_SMTP_PORT = 25;

	// ----------------------------------------------------- PRIVATE ATTRIBUTES
	/**
	 * Contains all the information exchanged between the SMTP server & the clients.
	 * <br/>It the SMTP server full LOG...
	 */
	private StringBuffer protocolExchangeLog = new StringBuffer(); 
	/**
	 * A reference to the thread running the current SMTP server.
	 */
	private Thread currentThread = null;
	
	// ------------------------------------------------------------- EXCEPTIONS
	// ------------------------------------------------------- INTERNAL CLASSES
	/**
	 * This class is used to process the SMTP request
	 */
	private class SmtpConnectionProcessor implements Runnable{
		
		// -------------------------------------------------------------- CONSTANTS
		private final static String RFC0821_CMD_CRLF = "\r\n";
		private final static String RFC0821_CMD_220_SERVERUP = "220 SMTP Server IS UP";
		private final static String RFC0821_CMD_HELO = "HELO";
		private final static String RFC0821_CMD_EHLO = "EHLO";		
		private final static String RCC0821_CMD_250_SERVERREADY = "250 SIMTREDCommons MockMail Server ready...";
		private final static String RFC0821_CMD_MAIL_FROM = "MAIL FROM:";
		private final static String RFC0821_CMD_250_OK = "250 OK";
		private final static String RFC0821_CMD_RCPT_TO = "RCPT TO:";
		private final static String RFC0821_CMD_DATA = "DATA";
		private final static String RFC0821_CMD_354_DATASTART = "354 Start mail input; end with <CRLF>.<CRLF>";
		private final static String RFC0821_CMD_ENDDATA = "\n.\n";
		private final static String RFC0821_CMD_QUIT = "QUIT";
		
		// ----------------------------------------------------- PRIVATE ATTRIBUTES
		
		/**
		 * Socket used to communicate with the client
		 */
	    private Socket socket;

	    /**
	     * Stream to the client.
	     */
	    private DataOutputStream toClient;

		/**
		 * Stream from the client.
		 */
	    private BufferedReader fromClient;
	    
	    /**
	     * A refrence to the SMTP Server LOG.
	     */
	    private StringBuffer serverLog;

		// ------------------------------------------------------------- EXCEPTIONS

		// ------------------------------------------------------- INTERNAL CLASSES

		// ----------------------------------------------------------- CONSTRUCTORS
	    
	    public SmtpConnectionProcessor(Socket socket, StringBuffer serverLog) throws Exception {
			this.socket = socket;
			this.serverLog = serverLog;
		}

		// --------------------------------------------------------- PUBLIC METHODS
	    
	    public void run() {
			try {
				processRequest();
			} catch (Exception e) {
				e.printStackTrace();
			} // end-try
		}

		// -------------------------------------------------------- PRIVATE METHODS
	    
	    /**
	     * Simulate an SMTP server and dump the client request to the console.
	     */
	    private void processRequest(){

			// flags to indicate client quit the connection
			boolean quit=false;
			
			// variable to store the client command
			String requestCommand;
			
			// Build the stream to & from the client
			{
				try {
					InputStream is = socket.getInputStream();
					toClient = new DataOutputStream(socket.getOutputStream());
	
					InputStreamReader sr = new InputStreamReader(is);
					fromClient = new BufferedReader(sr);
				}catch (IOException e) {
					e.printStackTrace();
				} // end-try
			} // end-block
			
			// Wait the client HELO command and reply HELO
			{
				reply(RFC0821_CMD_220_SERVERUP);
				
				while (!quit) {
					requestCommand=fetch();
					if ( (requestCommand).length()==0 ) quit=true;
					if (isCommand(requestCommand, RFC0821_CMD_HELO)) {
						break;
					} // end-if
					if (isCommand(requestCommand, RFC0821_CMD_EHLO)) {
						break;
					} // end-if
				} // end-while
				reply(RCC0821_CMD_250_SERVERREADY);
			} // end-while

			boolean dataTransfert = false;
			String allData = "";
			
			// Retrieve the mail & display it to the console.
			while (!quit) {
				requestCommand=fetch();
				
				// If the dataTransfert occurs, log all the data
				if (dataTransfert) {
					allData+= requestCommand + "\n";
				} // end-if
				
				if (requestCommand.startsWith(RFC0821_CMD_MAIL_FROM)) {
					reply(RFC0821_CMD_250_OK);
				} else if (requestCommand.startsWith("RCPT TO:")) {
					reply(RFC0821_CMD_250_OK);
				} else if (requestCommand.startsWith("DATA")) {
					reply(RFC0821_CMD_354_DATASTART);
					dataTransfert=true;
				} else if (dataTransfert && allData.endsWith(RFC0821_CMD_ENDDATA)) {
					reply(RFC0821_CMD_250_OK);
		 		} else {
		 			// Unkown command...
		 		} // end-if
				
				if ( null == requestCommand ) {
					quit=true;
				} // end-if
				if (isCommand(requestCommand, RFC0821_CMD_QUIT)) {
					quit=true;
				} // end-if
			} // end-while

			try {
				socket.close();
			}catch (IOException e) {
				e.printStackTrace();
			} // end-try
		}

	    /**
	     * Fetch the smtpClient command line.
	     */
		private String fetch(){
			String message="";
			try {
				message = fromClient.readLine();
			}catch (IOException e) {
				e.printStackTrace();
			} // end-try
			
			try {
				if (message == null) {
					socket.close();
					return "";
				} // end-if
			}catch (IOException e) {
				e.printStackTrace();
			} // end-try
			
			if (null != message) {
				System.out.println(message);
				serverLog.append(message + "\n");
			} // end-if
			
			return message;
		}

		/**
		 * Send the command to the client.
		 */
		private void reply (String command){
			try {
				if (!socket.isClosed()) toClient.writeBytes(command+RFC0821_CMD_CRLF);			
			} catch (IOException e) {
				e.printStackTrace();
			} // end-try
			System.out.println(command);
			serverLog.append(command + "\n");
			return;
		}
		
		/**
		 * @return true if the given command is stored in the smtpClientCommand.
		 */
		private boolean isCommand(String smtpClientCommand, String command) {
			if (smtpClientCommand.length()>=command.length() && smtpClientCommand.substring(0,command.length()).equals(command)) {
				return true;
			} // end-if
			return false;
		}

		// ------------------------------------------------------ PROTECTED METHODS
		
		/**
		 * When the connection with the client is ended, close the socket. 
		 */
	    protected void finalize() throws Throwable {
		    socket.close();
			super.finalize();
	    }

		// --------------------------------------------- ADVANCED GETTERS & SETTERS

		// ------------------------------------------------------ GETTERS & SETTERS

	}

	// ----------------------------------------------------------- CONSTRUCTORS
	
	/**
	 * Start/run a new SMTP Server with the default SMTP Port Number (25).
	 * @throws IOException
	 */
	public FakeSMTPServer() throws IOException {
		this(DEFAULT_SMTP_PORT);
	}
	
	/**
	 * Start/run a new SMTP Server with a custom SMTP Port Number. 
	 * @throws IOException
	 */
	public FakeSMTPServer(int port) throws IOException {
		super(port);
		currentThread = new Thread(this);
		currentThread.start();
	}

	// --------------------------------------------------------- PUBLIC METHODS
	/**
	 * @return the content of the current SMTP exchange LOG contains the given keyword to find.
	 */
	public final boolean isInLog(String keywordToFind) {
		return (this.protocolExchangeLog.toString().indexOf(keywordToFind)!=-1);
	}
	
	/**
	 * @return the content of the current SMTP exchange LOG
	 */
	public final String getLog() {
		return this.protocolExchangeLog.toString();
	}

	/**
	 * Method run
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		while (true) {
			try {
				Socket SMTPSocket = accept();
				SmtpConnectionProcessor connectionProcessor = new SmtpConnectionProcessor(SMTPSocket, protocolExchangeLog);
				Thread t = new Thread(connectionProcessor);
				t.start();
			} catch (Exception e) {
				e.printStackTrace();
			} // end-try
		} // end-while
	}
	
	// -------------------------------------------------------- PRIVATE METHODS
	// ------------------------------------------------------ PROTECTED METHODS
	/**
	 * Method finalize
	 * @throws Throwable
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}
	
	public static void main(String[] args) {
		try {
			FakeSMTPServer srv = new FakeSMTPServer();
			while (true) {
				Thread.sleep(10000);
			} // end-while
		} catch (Exception e) {
			e.printStackTrace();
		} // end-try
	}
	
	// --------------------------------------------- ADVANCED GETTERS & SETTERS
	// ------------------------------------------------------ GETTERS & SETTERS
}
